import inspect
import logging
from datetime import datetime
from functools import wraps
from hashlib import shake_256
from pathlib import Path
from pprint import pformat
from textwrap import dedent, fill, indent
from typing import Any, Callable

from graphviz import Digraph

current_directory = Path(__file__).resolve().parent
run_directory = (
    current_directory / f"runs/run-{datetime.now().strftime('%Y-%m-%d %H-%M-%S')}"
)
run_directory.mkdir(parents=True, exist_ok=False)
logger = logging.getLogger("test-classifier")
logger.setLevel(level=logging.DEBUG)
file_handler = logging.FileHandler(run_directory / "output.log")
file_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
logger.addHandler(file_handler)


class log_calls:
    description: str

    prolog = dedent(
        """
    ----------------------------------------
    {description}
    {func_name}:
        \tArguments:
        {arguments}
        """
    )

    epilog = dedent(
        """\tResult:
        {result}
    ----------------------------------------
    """
    )

    def __init__(self, description: str, level: int = logging.DEBUG):
        self.description = description
        self.level = level

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            arguments = self._get_named_arguments(func, *args, **kwargs)
            logger.log(
                self.level,
                self.prolog.format(
                    description=self.description,
                    func_name=func.__name__,
                    arguments=indent(pformat(arguments), "\t"),
                ),
            )

            result = func(*args, **kwargs)

            logger.log(
                self.level,
                self.epilog.format(
                    result=indent(pformat(result), "\t"),
                ),
            )
            return result

        return wrapper

    def _get_named_arguments(self, func: Callable[..., Any], *args, **kwargs):
        signature = inspect.signature(func)
        arguments = {}
        for argument_index, (argument_name, argument) in enumerate(signature.parameters.items()):
            if argument_index < len(args):
                # argument_name is from args
                value = args[argument_index]
            elif argument_name in kwargs:
                # argument_name is from kwargs
                value = kwargs[argument_name]
            else:
                # argument_name is from defaults
                value = argument.default

            arguments[argument_name] = value
        return arguments


@log_calls("Saving genealogy")
def save_genealogy(family_tree, P, S, T):
    dot = Digraph(comment="Genealogy")
    for t in range(T + 1):
        with dot.subgraph(name=f"cluster_{t}") as cluster:
            cluster.attr(
                label=f"Generation {t}",
                color="lightgrey",
            )
            for n in range(len(P[t])):
                cluster.node(f"{t},{n}", label=f"{fill(P[t][n])}\nAccuracy: {S[t][n]}")
                if parents := family_tree.get(P[t][n]):
                    if parents[0] in P[t]:
                        parent1_index = P[t].index(parents[0])
                    elif parents[0] in P[t - 1]:
                        parent1_index = P[t - 1].index(parents[0])
                    else:
                        logger.warning(
                            f"Parent `{parents[0]}` not found neither P[{t}] nor P[{t - 1}]"
                        )
                        continue

                    if parents[1] in P[t]:
                        parent2_index = P[t].index(parents[1])
                    elif parents[1] in P[t - 1]:
                        parent2_index = P[t - 1].index(parents[1])
                    else:
                        logger.warning(
                            f"Parent `{parents[1]}` not found neither P[{t}] nor P[{t - 1}]"
                        )
                        continue

                    dot.edge(f"{t},{parent1_index}", f"{t},{n}")
                    dot.edge(f"{t},{parent2_index}", f"{t},{n}")
                if t > 0 and P[t][n] in P[t - 1]:
                    parent = P[t - 1].index(P[t][n])
                    dot.edge(f"{t - 1},{parent}", f"{t},{n}", color="lightgrey")

    dot.render(run_directory / "genealogy", engine="dot")


def save_family_tree_visualization(family_tree, P, S, T):
    dot = Digraph(comment="Genealogy")
    node_id = lambda prompt: shake_256(prompt.encode("utf-8")).hexdigest(5)
    flatten = lambda l: [item for sublist in l for item in sublist]
    prompt_generations = lambda prompt: [str(t) for t in range(T) if prompt in P[t]]
    scores = {
        prompt: score
        for prompt, score in zip(
            flatten(P),
            flatten(S),
        )
    }
    node_template = dedent(
        """
        {prompt}
        ∈ P{{{prompt_generations}}}
        Accuracy: {score}"""
    )
    for prompt in family_tree:
        node_text = node_template.format(
            prompt=fill(prompt),
            prompt_generations=", ".join(prompt_generations(prompt)),
            score=scores[prompt],
        )
        dot.node(
            f"{node_id(prompt)}",
            label=node_text,
        )
    for prompt, parents in family_tree.items():
        if parents is not None:
            # TODO consider more than 2 parents?
            dot.edge(
                f"{node_id(parents[0])}",
                f"{node_id(prompt)}",
            )
            dot.edge(
                f"{node_id(parents[1])}",
                f"{node_id(prompt)}",
            )

    dot.render(run_directory / "family_tree", engine="dot")


def save_snapshot(family_tree, P, S, T, N):
    import json

    with open(run_directory / "snapshot.json", "w") as f:
        json.dump(
            {
                "family_tree": family_tree,
                "P": P,
                "S": S,
                "T": T,
                "N": N,
            },
            f,
            indent=4,
        )


def load_snapshot(path: Path):
    import json

    with path.open("r") as f:
        snapshot = json.load(f)
    return (
        snapshot["family_tree"],
        snapshot["P"],
        snapshot["S"],
        snapshot["T"],
        snapshot["N"],
    )
